import decimal
from datetime import datetime

from easytradesdk.Serializer import DeserializableObject


class Ticker(DeserializableObject):

    def __init__(self, price=None, lastUpdateMills=0):
        self.price = price
        self.lastUpdateMills = lastUpdateMills

    def getObjectMapper(self):
        return {"price": decimal.Decimal}

    def getOffsetMills(self):
        _currentMills = int(round(datetime.now().timestamp() * 1000))
        return _currentMills - self.lastUpdateMills



import abc

class AbsInstance(metaclass=abc.ABCMeta):

    # 实例初始化
    @abc.abstractmethod
    def init(self):
        pass

    # 实例销毁
    @abc.abstractmethod
    def destroy(self):
        pass

    # 实例执行函数，子类需要重写
    @abc.abstractmethod
    def execute(self):
        pass

    @abc.abstractmethod
    def executeStopLoss(self):
        pass

    @abc.abstractmethod
    def executeStopProfit(self):
        pass



import abc


class AbsMarketApi(metaclass=abc.ABCMeta):

    def fetchKline(self, tc, symbol, klinePeriod, startTimeMills=None, endTimeMills=None, limit=200):
        pass

    def fetchLatestKline(self, tc, symbol, klinePeriod, endTimeMills=None, limit=200, excludeCurrent=None):
        pass

    def fetchDepth(self, tc, symbol, depthLimit=20):
        pass

    def fetchTicker(self, tc, symbol):
        pass



from datetime import datetime

from easytradesdk.Serializer import DeserializableObject


class Condition(DeserializableObject):

    def __init__(self, identify: str = None, conditionData: dict = None):
        self.__identify = identify  # 开仓/卖出条件的标识
        self.__timeStamp = int(round(datetime.now().timestamp() * 1000))  # 开仓/卖出条件产生的时间
        self.__conditionData = conditionData  # 开仓/卖出条件产生时的条件数据
        self.__canceled = False

    def getObjectMapper(self):
        return {"__conditionData": dict}

    def cancel(self):
        self.__canceled = True

    def isCanceled(self):
        return self.__canceled

    def getIdentify(self):
        return self.__identify

    def generateBuyOrderSignal(self, signalData: dict = None):
        return OrderSignal(self, 1, signalData)

    def generateSellOrderSignal(self, signalData: dict = None):
        return OrderSignal(self, -1, signalData)


class OrderSignal(DeserializableObject):

    def __init__(self, condition: Condition = None, signalType: int = None, signalData: dict = None):
        self.__condition = condition
        self.__signalType = signalType  # 1 买入 -1 卖出
        self.__timeStamp = int(round(datetime.now().timestamp() * 1000))  # 买入或卖出信号产生的时间
        self.__signalData = signalData  # 买入或卖出信号产生时的信号数据
        self.__isSignalExecuted = False  # 信号是否已经执行

    def getObjectMapper(self):
        return {"__condition": Condition, "__signalData": dict}

    def isExecuted(self):
        return self.__isSignalExecuted

    def executedCallBack(self, result: True):
        self.__isSignalExecuted = result

    def getCondition(self):
        return self.__condition

    def getSignalType(self):
        return self.__signalType

    def getTimeStamp(self):
        return self.__timeStamp

    def getSignalData(self):
        return self.__signalData



from easytradesdk.Serializer import DeserializableObject
from easytradesdk.support.Condition import Condition


class Conditions(DeserializableObject):
    __DEFAULT_IDENTIFY = "default"

    def __init__(self):
        self.__conditions = []

    def getObjectMapper(self):
        return {"__conditions", Condition}

    def exists(self, identify: str = __DEFAULT_IDENTIFY):
        return self.get(identify) is not None

    def get(self, identify: str = __DEFAULT_IDENTIFY):
        if self.__conditions:
            for _condition in self.__conditions:
                if _condition.getIdentify() == identify:
                    return _condition
        return None

    def getFirst(self):
        if self.__conditions:
            return self.__conditions[0]
        return None

    def getLast(self):
        if self.__conditions:
            return self.__conditions[len(self.__conditions) - 1]

    def getAll(self):
        return self.__conditions

    def length(self):
        return len(self.__conditions)

    def clear(self):
        for _condition in self.__conditions:
            if _condition is not None:
                _condition.cancel()
        self.__conditions.clear()

    def remove(self, identify: str = __DEFAULT_IDENTIFY):
        _idxs = []
        for _index, _condition in enumerate(self.__conditions):
            if _condition.getIdentify() == identify:
                _idxs.append(_index)

        for _i in _idxs:
            del self.__conditions[_i]

    def cancel(self, identify: str = __DEFAULT_IDENTIFY):
        _condition = self.get(identify)
        if _condition:
            _condition.cancel()

    def cancelAll(self):
        for _condition in self.__conditions:
            _condition.cancel()

    def generate(self, conditionData: dict = None, identify: str = __DEFAULT_IDENTIFY):
        _condition = Condition(identify, conditionData)
        self.__conditions.append(_condition)
        return _condition

    def append(self, condition: Condition):
        if condition:
            self.__conditions.append(condition)



from easytradesdk.indicator.SimpleIndicator import SimpleIndicator
import talib._ta_lib as ta


class Cci(SimpleIndicator):

    def __init__(self):
        super(Cci, self).__init__()

    @staticmethod
    def generate(np_high_price_array, np_low_price_array, np_closing_price_array, time_periods):

        if (np_high_price_array is None or len(np_high_price_array) == 0) or \
                (np_low_price_array is None or len(np_low_price_array) == 0) or \
                (np_closing_price_array is None or len(np_closing_price_array) == 0):
            return []

        _cci_tmp_dict = {}
        for _t in range(0, len(time_periods)):
            _time_period = time_periods[_t]
            _cci_tmp_dict[_time_period] = ta.CCI(np_high_price_array, np_low_price_array, np_closing_price_array, timeperiod=_time_period)
        _cci_list = []
        for _index in range(0, len(np_closing_price_array)):
            _cci = Cci()
            for t in range(0, len(time_periods)):
                _time_period = time_periods[t]
                _np_cci_array = _cci_tmp_dict[_time_period]
                _cci_value = _np_cci_array[_index]

                if _cci_value is not None:
                    _cci.setValue(_cci_value, _time_period)
                else:
                    _cci.setValue(None, _time_period)
            _cci_list.append(_cci)

        return _cci_list



from easytradesdk.indicator.SimpleIndicator import SimpleIndicator
import talib._ta_lib as ta


class Ma(SimpleIndicator):

    def __init__(self):
        super(Ma, self).__init__()

    @staticmethod
    def generate(np_closing_price_array, time_periods):

        if np_closing_price_array is None or len(np_closing_price_array) == 0:
            return []

        _ma_dict = {}
        for _t in range(0, len(time_periods)):
            _time_period = time_periods[_t]
            _ma_dict[_time_period] = ta.MA(np_closing_price_array, timeperiod=_time_period, matype=0)

        _ma_list = []
        for _index in range(0, len(np_closing_price_array)):
            _ma = Ma()
            for t in range(0, len(time_periods)):
                _time_period = time_periods[t]
                _np_ma_array = _ma_dict[_time_period]
                _v = _np_ma_array[_index]

                if _v is not None:
                    _ma.setValue(_v, _time_period)
                else:
                    _ma.setValue(None, _time_period)
            _ma_list.append(_ma)

        return _ma_list



def singleton(cls):
    _instance_dict = {}

    def inner(*args, **kwargs):
        if cls not in _instance_dict:
            _instance_dict[cls] = cls(*args, **kwargs)
        return _instance_dict.get(cls)

    return inner



import abc
import decimal
from easytradesdk.support.Condition import OrderSignal


class AbsTradeApi(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def getOrder(self, clientOrderId):
        """
            根据用户订单ID获取订单信息
            :param clientOrderId: 订单编号
            :return:
        """
        pass

    @abc.abstractmethod
    def queryOrders(self, startTimeMills=None, endTimeMills=None, tc=None, symbol=None, instanceId=None, limit=50):
        """
            根据订单时间戳查询订单信息，默认查询50条数据
            :param tc:              交易所
            :param symbol:          交易标的
            :param startTimeMills:  开始时间（下单时间）
            :param endTimeMills:    结束时间戳（下单时间）
            :param instanceId:      实例编号
            :param limit: default = 50
            :return:
        """
        pass

    # 下市价买单
    @abc.abstractmethod
    def buyMarketOrder(self, tc, symbol, quantity: decimal.Decimal, qtyScale=2, orderSignal: OrderSignal = None, orderData: dict = None):
        pass

    # 下市价卖单
    @abc.abstractmethod
    def sellMarketOrder(self, tc, symbol, quantity: decimal.Decimal, qtyScale=2, orderSignal: OrderSignal = None, orderData: dict = None):
        pass

    # 限价买单
    @abc.abstractmethod
    def buyLimitOrder(self, tc, symbol, price: decimal.Decimal, quantity: decimal.Decimal, qtyScale=2, orderSignal: OrderSignal = None, orderData: dict = None):
        pass

    # 限价卖单
    @abc.abstractmethod
    def sellLimitOrder(self, tc, symbol, price: decimal.Decimal, quantity: decimal.Decimal, qtyScale=2, orderSignal: OrderSignal = None, orderData: dict = None):
        pass



import datetime
import decimal

from easytradesdk.Serializer import DeserializableObject


class Kline(DeserializableObject):

    def __init__(self):
        self.tc = None
        self.symbol = None
        self.timeInterval = None
        self.timeMills = None
        self.endTimeMills = None
        self.startTime = None
        self.endTime = None
        self.openingPrice = None
        self.closingPrice = None
        self.highPrice = None
        self.lowPrice = None
        self.volume = None
        self.quantity = None
        self.cnt = None
        self.buyVolume = None
        self.buyQuantity = None

    def getObjectMapper(self):
        return {
            "startTime": datetime.datetime, "endTime": datetime.datetime,
            "openingPrice": decimal.Decimal, "closingPrice": decimal.Decimal,
            "highPrice": decimal.Decimal, "lowPrice": decimal.Decimal,
            "volume": decimal.Decimal, "quantity": decimal.Decimal,
            "buyVolume": decimal.Decimal, "buyQuantity": decimal.Decimal}

    def isFall(self):
        return self.closingPrice < self.openingPrice

    def isRise(self):
        return self.closingPrice > self.openingPrice

    @staticmethod
    def getCount(startTimeMills, endTimeMills, interval):

        symbol = interval[-1]
        cnt = int(interval[0:len(interval) - 1])

        if symbol == 'm':
            return int((startTimeMills - endTimeMills) / (cnt * 60 * 1000))

        if symbol == 'h':
            return int((startTimeMills - endTimeMills) / (cnt * 3600 * 1000))

        if symbol == 'd':
            return int((startTimeMills - endTimeMills) / (cnt * 3600 * 24 * 1000))

        return 0



from easytradesdk.indicator.SimpleIndicator import SimpleIndicator
import talib._ta_lib as ta


class Wma(SimpleIndicator):

    def __init__(self):
        super(Wma, self).__init__()

    @staticmethod
    def generate(np_closing_price_array, time_periods):

        if np_closing_price_array is None or len(np_closing_price_array) == 0:
            return []

        _wma_dict = {}
        for _t in range(0, len(time_periods)):
            _time_period = time_periods[_t]
            _wma_dict[_time_period] = ta.WMA(np_closing_price_array, timeperiod=_time_period)

        _wma_list = []
        for _index in range(0, len(np_closing_price_array)):
            _wma = Wma()
            for t in range(0, len(time_periods)):
                _time_period = time_periods[t]
                _np_wma_array = _wma_dict[_time_period]
                _v = _np_wma_array[_index]

                if _v is not None:
                    _wma.setValue(_v, _time_period)
                else:
                    _wma.setValue(None, _time_period)
            _wma_list.append(_wma)

        return _wma_list



import decimal
from datetime import datetime

from easytradesdk.Serializer import DeserializableObject


class DepthValue(DeserializableObject):

    def __init__(self):
        self.price = None
        self.quantity = None

    def getObjectMapper(self):
        return {"price": decimal.Decimal, "quantity": decimal.Decimal}


class Depth(DeserializableObject):

    def __init__(self, bids=None, asks=None, lastUpdateMills=0):
        if bids is None:
            bids = []
        if asks is None:
            asks = []
        self.bids = bids
        self.asks = asks
        self.lastUpdateMills = lastUpdateMills
        self.bids = []
        self.asks = []

    def getObjectMapper(self):
        return {"bids": DepthValue, "asks": DepthValue}

    def getOffsetMills(self):
        _currentMills = int(round(datetime.now().timestamp() * 1000))
        return _currentMills - self.lastUpdateMills



import decimal

from easytradesdk.Serializer import DeserializableObject


class Position(DeserializableObject):

    def __init__(self):
        self.tc = None
        self.symbol = None
        self.initialTicker = decimal.Decimal('0')
        self.initialAmount = decimal.Decimal('0')
        self.initialHolding = decimal.Decimal('0')
        self.initialTotalAmount = decimal.Decimal('0')
        self.remainAmount = decimal.Decimal('0')
        self.remainHolding = decimal.Decimal('0')
        self.lastDealAmount = decimal.Decimal('0')
        self.lastDealQuantity = decimal.Decimal('0')
        self.lastTicker = decimal.Decimal('0')

    def getObjectMapper(self):
        return {"initialTicker": decimal.Decimal, "initialAmount": decimal.Decimal,
                "initialHolding": decimal.Decimal, "initialTotalAmount": decimal.Decimal,
                "remainAmount": decimal.Decimal, "remainHolding": decimal.Decimal,
                "lastDealAmount": decimal.Decimal, "lastDealQuantity": decimal.Decimal, "lastTicker": decimal.Decimal}



import abc


class AbsContext(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def getPosition(self, tc, symbol):
        pass

    @abc.abstractmethod
    def getPositions(self):
        pass

    @abc.abstractmethod
    def getStrategyParams(self):
        pass

    @abc.abstractmethod
    def getExecutePeriod(self):
        pass

    @abc.abstractmethod
    def getExecutingTimeMills(self):
        pass

    @abc.abstractmethod
    def getMarketApi(self):
        pass

    @abc.abstractmethod
    def getTradeApi(self):
        pass

    @abc.abstractmethod
    def calculateMaxAvailableBuyQuantity(self, tc, symbol, amount=None, offsetPrice=None, scale=2):
        pass



import abc


class AbsInstance(metaclass=abc.ABCMeta):

    # 实例初始化
    @abc.abstractmethod
    def init(self, context):
        pass

    # 实例销毁
    @abc.abstractmethod
    def destroy(self, context):
        pass

    # 实例执行函数，子类需要重写
    @abc.abstractmethod
    def execute(self, context):
        pass

    @abc.abstractmethod
    def executeStopLoss(self, context):
        pass

    @abc.abstractmethod
    def executeStopProfit(self, context):
        pass


class Instance(AbsInstance):

    def __init__(self):
        # 延时执行, 默认6秒
        self.executeDelaySeconds = 6

    def init(self, context):
        pass

    # 实例销毁
    def destroy(self, context):
        pass

    # 实例执行函数，子类需要重写
    def execute(self, context):
        pass

    # 默认的止损算法，子类可以重写
    def executeStopLoss(self, context):
        pass

    # 执行止盈操作, 子类可以重写
    def executeStopProfit(self, context):
        pass



from easytradesdk.indicator.SimpleIndicator import SimpleIndicator
import talib._ta_lib as ta


class Ema(SimpleIndicator):

    def __init__(self):
        super(Ema, self).__init__()

    @staticmethod
    def generate(np_closing_price_array, time_periods):

        if np_closing_price_array is None or len(np_closing_price_array) == 0:
            return []

        _ema_dict = {}
        for _t in range(0, len(time_periods)):
            _time_period = time_periods[_t]
            _ema_dict[_time_period] = ta.EMA(np_closing_price_array, timeperiod=_time_period)

        _ema_list = []
        for _index in range(0, len(np_closing_price_array)):
            _ema = Ema()
            for t in range(0, len(time_periods)):
                _time_period = time_periods[t]
                _np_ema_array = _ema_dict[_time_period]
                _v = _np_ema_array[_index]

                if _v is not None:
                    _ema.setValue(_v, _time_period)
                else:
                    _ema.setValue(None, _time_period)
            _ema_list.append(_ema)

        return _ema_list



import datetime
import decimal

from easytradesdk.Serializer import DeserializableObject
from easytradesdk.entity.Position import Position
from easytradesdk.support.Condition import OrderSignal


class Order(DeserializableObject):

    def __init__(self):
        self.instanceId = None  # 实例编号
        self.strategyInstanceType = None  # 实例类型
        self.tc = None  # 交易所
        self.symbol = None  # 交易标的
        self.orderId = None # 交易所订单编号
        self.clientOrderId = None  # 生成的客户端订单ID
        self.price = None  # 买卖价格, 市价单价格为0
        self.totalQty = None  # 下单量
        self.dealQty = None  # 实际成交量
        self.type = None  # limit, market
        self.side = None  # buy, sell
        self.status = None  # new, partially_dealed, dealed, rejected, expired
        self.time = None
        self.timeMills = None
        self.updateTimeMills = None
        self.totalDealAmount = None
        self.totalFeeAmount = None  # 手续费金额
        self.feeSymbol = None  # 手续费标的
        self.dealDetails = None  # 手续费标的
        self.strategyParams = None  # 下单时的策略参数
        self.orderData = None  # 用户订单数据
        self.orderSignal = None  # 订单信号数据
        self.actualDealQty = None  # 扣除手续费之后的实际成交量
        self.actualDealAmount = None  # 扣除手续费之后的实际成交额
        self.posBeforeDeal = None  # 交易前那一刻的仓位, 仅实盘有值
        self.posAfterDeal = None  # 交易后那一刻的仓位, 仅实盘有值
        self.positions = None

    def getObjectMapper(self):
        return {
            "time": datetime.datetime,
            "price": decimal.Decimal, "totalQty": decimal.Decimal,
            "dealQty": decimal.Decimal, "actualDealQty": decimal.Decimal,
            "actualDealAmount": decimal.Decimal, "totalFeeAmount": decimal.Decimal, "totalDealAmount": decimal.Decimal,
            "strategyParams": dict, "orderData": dict, "orderSignal": OrderSignal, "positions": Position}



from easytradesdk.Serializer import DeserializableObject


class SimpleIndicator(DeserializableObject):

    def __init__(self):
        self.value = None

    def getObjectMapper(self):
        return {"value": dict}

    def setValue(self, value, time_period):
        if self.value is None:
            self.value = {time_period: value}
        else:
            self.value[time_period] = value

    def getValue(self, time_period):
        return self.value[time_period]

    def getWholeValue(self):
        return self.value



'''
    模拟交易或者回测时需要设置的滑点
'''
import decimal

from easytradesdk.Serializer import DeserializableObject


class Slippage(DeserializableObject):

    def __init__(self, slippageType=1, value=0):

        # 滑点类型 1:百分比  2:某个具体的值
        self.slippageType = slippageType

        self.value = decimal.Decimal(str(value))

    def getObjectMapper(self):
        return {"value": decimal.Decimal}



class TCConst:

    def __init__(self):
        self.__BINANCE = "BINANCE"

    @property
    def BINANCE(self):
        return self.__BINANCE


class SymbolConst:

    def __init__(self):
        self.__BTC_USDT = "BTC_USDT"
        self.__ETH_BTC = "ETH_BTC"
        self.__ETH_USDT = "ETH_USDT"

    @property
    def BTC_USDT(self):
        return self.__BTC_USDT

    @property
    def ETH_BTC(self):
        return self.__ETH_BTC

    @property
    def ETH_USDT(self):
        return self.__ETH_USDT


class KlinePeriodConst:

    def __init__(self):
        self.__MIN_1 = "1m"
        self.__MIN_3 = "3m"
        self.__MIN_5 = "5m"
        self.__MIN_15 = "15m"
        self.__MIN_30 = "30m"
        self.__HOUR_1 = "1h"
        self.__HOUR_2 = "2h"
        self.__HOUR_4 = "4h"
        self.__HOUR_6 = "6h"
        self.__HOUR_8 = "8h"
        self.__HOUR_12 = "12h"
        self.__DAY_1 = "1d"
        self.__DAY_3 = "3d"
        self.__WEEK_1 = "1w"
        self.__MONTH_1 = "1M"

    @property
    def MIN_1(self):
        return self.__MIN_1

    @property
    def MIN_3(self):
        return self.__MIN_3

    @property
    def MIN_5(self):
        return self.__MIN_5

    @property
    def MIN_15(self):
        return self.__MIN_15

    @property
    def MIN_30(self):
        return self.__MIN_30

    @property
    def HOUR_1(self):
        return self.__HOUR_1

    @property
    def HOUR_2(self):
        return self.__HOUR_2

    @property
    def HOUR_4(self):
        return self.__HOUR_4

    @property
    def HOUR_6(self):
        return self.__HOUR_6

    @property
    def HOUR_8(self):
        return self.__HOUR_8

    @property
    def HOUR_12(self):
        return self.__HOUR_12

    @property
    def DAY_1(self):
        return self.__DAY_1

    @property
    def DAY_3(self):
        return self.__DAY_3

    @property
    def WEEK_1(self):
        return self.__WEEK_1

    @property
    def MONTH_1(self):
        return self.__MONTH_1


TC = TCConst()
Symbol = SymbolConst()
KlinePeriod = KlinePeriodConst()


import numpy as np
import talib._ta_lib as ta


class BBI:

    @staticmethod
    def generate(np_closing_price_array, period1, period2, period3, period4):

        _max_period = max(period1, period2, period3, period4)
        if len(np_closing_price_array) < _max_period:
            return []

        _result = []
        _ma1_array = ta.MA(np_closing_price_array, period1)
        _ma2_array = ta.MA(np_closing_price_array, period2)
        _ma3_array = ta.MA(np_closing_price_array, period3)
        _ma4_array = ta.MA(np_closing_price_array, period4)

        for i, v in enumerate(np_closing_price_array):
            _ma1 = _ma1_array[i]
            _ma2 = _ma2_array[i]
            _ma3 = _ma3_array[i]
            _ma4 = _ma4_array[i]
            if np.isnan(_ma1) or np.isnan(_ma2) or np.isnan(_ma3) or np.isnan(_ma4):
                _result.append(np.NAN)
            else:
                _result.append((_ma1 + _ma2 + _ma3 + _ma4) / 4)

        return _result



import talib._ta_lib as ta
import numpy as np


class Kdj:

    @staticmethod
    def generate(np_high_array, np_low_array, np_close_array, n, m1, m2):

        _np_kdj_array = ta.STOCH(np_high_array,
                                 np_low_array,
                                 np_close_array,
                                 fastk_period=n,
                                 slowk_period=2 * m1 - 1,
                                 slowk_matype=1,
                                 slowd_period=2 * m2 - 1,
                                 slowd_matype=1)

        _result = []

        for _index in range(0, len(np_close_array)):

            _k = _np_kdj_array[0][_index]
            _d = _np_kdj_array[1][_index]

            if np.isnan(_k) or np.isnan(_d):
                _result.append(None)
            else:
                _j = 3 * _k - 2 * _d
                _result.append({"k": _k, "d": _d, "j": _j})

        return _result



from easytradesdk.indicator.SimpleIndicator import SimpleIndicator
import talib._ta_lib as ta


class Tema(SimpleIndicator):

    def __init__(self):
        super(Tema, self).__init__()

    @staticmethod
    def generate(np_closing_price_array, time_periods):

        if (np_closing_price_array is None or len(np_closing_price_array) == 0):
            return []

        _tema_tmp_dict = {}
        for _t in range(0, len(time_periods)):
            _time_period = time_periods[_t]
            _tema_tmp_dict[_time_period] = ta.TEMA(np_closing_price_array, timeperiod=_time_period)
        _tema_list = []
        for _index in range(0, len(np_closing_price_array)):
            _tema = Tema()
            for t in range(0, len(time_periods)):
                _time_period = time_periods[t]
                _np_tema_array = _tema_tmp_dict[_time_period]
                _tema_value = _np_tema_array[_index]

                if _tema_value is not None:
                    _tema.setValue(_tema_value, _time_period)
                else:
                    _tema.setValue(None, _time_period)
            _tema_list.append(_tema)

        return _tema_list



import abc
import decimal
import json
from datetime import datetime, date


class DefaultJsonEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return str(o)
        if isinstance(o, datetime):
            return o.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(o, date):
            return o.strftime('%Y-%m-%d')
        return o.__dict__


class DeserializableObject(metaclass=abc.ABCMeta):

    def getObjectMapper(self):
        return {}


def objectToJson(obj):
    """
        python 对象转 json 字符串
        :param obj:
        :return:
    """
    if obj is not None:
        return json.dumps(obj, cls=DefaultJsonEncoder)
    return None


def jsonToObject(jsonStr, objClass):
    """
        json 反序列化 python object, objClass 需要继承 DeserializableObject
        :param jsonStr:
        :param objClass:
        :return: objClass 不继承 DeserializableObject 则返回 dict 类型，否则返回 objClass 类型
    """
    if jsonStr:
        dictData = json.loads(jsonStr, parse_float=decimal.Decimal)
        if not issubclass(objClass, DeserializableObject):
            return dictData
        obj = objClass()
        __setAttrFromDict(obj, dictData)
        return __dictToObject0(obj)
    return None


def jsonToDict(jsonStr):
    """
        json 字符串转 python dict
        :param jsonStr:
        :return:
    """
    if jsonStr:
        return json.loads(jsonStr)
    return None


def jsonArrayToObjectList(jsonStr, objClass):
    """
        json 字符串转 python object list
        :param jsonStr:
        :param objClass:
        :return: 如果 objClass 未继承 DeserializableObject, 则默认返回 dict list
    """
    if jsonStr:
        _list = []
        _dataList = json.loads(jsonStr)
        if not isinstance(_dataList, list):
            raise Exception("json string is not json array format")
        return dictListToObjectList(_dataList, objClass)
    return []


def dictToObject(dictData, objClass):
    """
        python dict 转 python object
        :param dictData:
        :param objClass:
        :return:如果 objClass
    """
    if objClass is None or dictData is None:
        return None

    if not isinstance(dictData, dict):
        raise Exception("type of dictData is not dict")

    if not issubclass(objClass, DeserializableObject):
        return dictData

    obj = objClass()
    __setAttrFromDict(obj, dictData)
    return __dictToObject0(obj)


def dictListToObjectList(dictList, objClass):
    """
        python dict list 转 object list
        :param dictList:
        :param objClass:
        :return:
    """
    if dictList is None:
        return None

    if len(dictList) == 0:
        return []

    if not isinstance(dictList, list):
        raise Exception("type of dictList is not list")

    _list = []
    for dictData in dictList:
        _list.append(dictToObject(dictData, objClass))
    return _list


def __dictToObject0(obj):
    if isinstance(obj, list):
        for _data in obj:
            if isinstance(_data, DeserializableObject):
                __dictToObject00(_data)

    elif isinstance(obj, DeserializableObject):
        __dictToObject00(obj)
    return obj


def __dictToObject00(obj):
    _dict_data = obj.__dict__
    _object_mapper = obj.getObjectMapper()

    for _p, _class in _object_mapper.items():

        _p_name = None
        _data = None
        _className = obj.__class__.__name__
        _private_key = "_" + _className + _p

        if _p in _dict_data:
            _p_name = _p
            _data = _dict_data[_p]
        elif _private_key in _dict_data:
            _p_name = _private_key
            _data = _dict_data[_private_key]

        if _data is None:
            continue

        if _class == dict:

            if isinstance(_data, str):
                try:
                    _data = json.loads(_data, parse_float=decimal.Decimal)
                except Exception as e:
                    print(e)

            if isinstance(_data, dict):
                setattr(obj, _p_name, _data)
                continue

            if isinstance(_data, list):
                _list = []
                for _d in _data:
                    if isinstance(_d, dict):
                        _list.append(_d)
                setattr(obj, _p_name, _list)
                continue

        if _class == decimal.Decimal:
            if isinstance(_data, str):
                setattr(obj, _p_name, decimal.Decimal(_data))
                continue

        if _class == datetime:
            if isinstance(_data, str):
                setattr(obj, _p_name, datetime.strptime(_data, "%Y-%m-%d %H:%M:%S"))
                continue

        if _class == date:
            if isinstance(_data, str):
                setattr(obj, _p_name, datetime.strptime(_data, "%Y-%m-%d").date())
                continue

        if not issubclass(_class, DeserializableObject):
            continue

        if isinstance(_data, str):
            try:
                _data = json.loads(_data, parse_float=decimal.Decimal)
            except Exception as e:
                print(e)

        if isinstance(_data, dict):
            __dictToObject0(__resolveDict(obj, _p_name, _class, _data))

        elif isinstance(_data, list):
            __dictToObject0(__resolveDictList(obj, _p_name, _class, _data))
        else:
            _o = _class()
            setattr(obj, _p_name, _o)
            __dictToObject0(_o)


def __resolveDict(obj, p_name, _class, dict_data):
    _o = _class()
    __setAttrFromDict(_o, dict_data)
    setattr(obj, p_name, _o)
    return _o


def __resolveDictList(obj, p_name, _class, dict_list):
    _l = []
    for _d in dict_list:
        _o = _class()
        if isinstance(_d, dict):
            __setAttrFromDict(_o, _d)
        _l.append(_o)
    setattr(obj, p_name, _l)
    return _l


def __setAttrFromDict(obj, dict_data):
    for _k, _v in dict_data.items():
        setattr(obj, _k, _v)



import decimal

import numpy as np
import talib._ta_lib as ta

from easytradesdk.Serializer import DeserializableObject


class Macd(DeserializableObject):

    def __init__(self, diff=None, dea=None, value=None):
        self.diff = diff
        self.dea = dea
        self.value = value

    def getObjectMapper(self):
        return {"diff": decimal.Decimal, "dea": decimal.Decimal, "value": decimal.Decimal}

    @staticmethod
    def generate(np_closing_price_array, fast_period, slow_period, signal_period):

        if np_closing_price_array is None or len(np_closing_price_array) == 0:
            return []

        _np_macd_array = ta.MACD(np_closing_price_array, fast_period, slow_period, signal_period)

        _macd_list = []

        for _index in range(0, len(np_closing_price_array)):
            _diff = _np_macd_array[0][_index]
            _dea = _np_macd_array[1][_index]
            _value = _np_macd_array[2][_index]

            if np.isnan(_diff) or np.isnan(_dea) or np.isnan(_value):
                _macd_list.append(None)
            else:
                _macd_list.append(Macd(_diff, _dea, _value))

        return _macd_list



from easytradesdk.indicator.SimpleIndicator import SimpleIndicator
import talib._ta_lib as ta


class Rsi(SimpleIndicator):

    def __init__(self):
        super(Rsi, self).__init__()

    @staticmethod
    def generate(np_closing_price_array, time_periods):
        if np_closing_price_array is None or len(np_closing_price_array) == 0:
            return []

        _rsi_tmp_dict = {}
        for t in range(0, len(time_periods)):
            _time_period = time_periods[t]

            _rsi_tmp_dict[_time_period] = ta.RSI(np_closing_price_array, timeperiod=_time_period)

        _rsi_list = []
        for _index in range(0, len(np_closing_price_array)):
            _rsi = Rsi()
            for _t in range(0, len(time_periods)):
                _time_period = time_periods[_t]
                _np_rsi_array = _rsi_tmp_dict[_time_period]
                _rsi_value = _np_rsi_array[_index]

                if _rsi_value is not None:
                    _rsi.setValue(_rsi_value, _time_period)
                else:
                    _rsi.setValue(None, _time_period)
            _rsi_list.append(_rsi)

        return _rsi_list



import numpy as np

from easytradesdk.indicator.SimpleIndicator import SimpleIndicator


class Vroc(SimpleIndicator):

    def __init__(self):
        super(Vroc, self).__init__()

    @staticmethod
    def generate(volume_array, time_periods):

        if (volume_array is None or len(volume_array) == 0):
            return []

        _vroc_tmp_dict = {}
        for _t in range(0, len(time_periods)):
            _time_period = time_periods[_t]
            _vroc_tmp_dict[_time_period] = Vroc.calculateVROC(volume_array, _time_period)

        _vroc_list = []
        for _index in range(0, len(volume_array)):
            _vroc = Vroc()
            for t in range(0, len(time_periods)):

                _time_period = time_periods[t]
                _np_vroc_array = _vroc_tmp_dict[_time_period]
                _vroc_value = _np_vroc_array[_index]

                if _vroc_value is not None:
                    _vroc.setValue(_vroc_value, _time_period)
                else:
                    _vroc.setValue(None, _time_period)
            _vroc_list.append(_vroc)

        return _vroc_list

    @staticmethod
    def calculateVROC(volume_array, period):

        if not volume_array or len(volume_array) < period:
            return []

        result = []
        for i, v in enumerate(volume_array):
            j = i - period
            if j < 0:
                result.append(np.NAN)
            else:
                if volume_array[j] == 0:
                    result.append(0)
                else:
                    result.append((volume_array[i] - volume_array[j]) / volume_array[j] * 100)
        return result



import decimal

import numpy as np
import talib._ta_lib as ta

from easytradesdk.Serializer import DeserializableObject


class Boll(DeserializableObject):

    def __init__(self, upper=None, middle=None, down=None):
        self.upper = upper
        self.middle = middle
        self.down = down

    def getObjectMapper(self):
        return {"upper": decimal.Decimal, "middle": decimal.Decimal, "down": decimal.Decimal}

    @staticmethod
    def generate(np_closing_price_array, time_period, nbdevup, nbdevdn, ma_type):

        if np_closing_price_array is None or len(np_closing_price_array) == 0:
            return []

        _np_boll_Array = ta.BBANDS(np_closing_price_array, time_period, nbdevup, nbdevdn, ma_type)

        _bill_list = []

        for index in range(0, len(np_closing_price_array)):
            _boll_upper = _np_boll_Array[0][index]
            _boll_middle = _np_boll_Array[1][index]
            _boll_down = _np_boll_Array[2][index]

            if np.isnan(_boll_upper) or np.isnan(_boll_middle) or np.isnan(_boll_down):
                _bill_list.append(None)
            else:
                _bill_list.append(Boll(_boll_upper, _boll_middle, _boll_down))

        return _bill_list

    @staticmethod
    def getBollMiddles(_bill_list):
        _result = []
        for i in range(0, len(_bill_list)):
            _result.append(_bill_list[i].middle)
        return _result

    @staticmethod
    def getBollUppers(_bill_list):
        _result = []
        for i in range(0, len(_bill_list)):
            _result.append(_bill_list[i].upper)
        return _result

    @staticmethod
    def getBollDowns(_bill_list):
        _result = []
        for i in range(0, len(_bill_list)):
            _result.append(_bill_list[i].down)
        return _result



import numpy as np


class KlineSupporter:

    def __init__(self, klines):

        self.__openingPriceList = []
        self.__closingPriceList = []
        self.__highPriceList = []
        self.__lowPriceList = []
        self.__volumeList = []
        self.__openingPriceNpArray = None
        self.__closingPriceNpArray = None
        self.__highPriceNpArray = None
        self.__lowPriceNpArray = None
        self.__volumeNpArray = None

        if klines is not None and len(klines) > 0:

            for idx in range(0, len(klines)):
                self.__openingPriceList.append(klines[idx].openingPrice)
                self.__closingPriceList.append(klines[idx].closingPrice)
                self.__highPriceList.append(klines[idx].highPrice)
                self.__lowPriceList.append(klines[idx].lowPrice)
                self.__volumeList.append(klines[idx].quantity)

            self.__openingPriceNpArray = np.array(self.__openingPriceList, dtype=np.double)
            self.__closingPriceNpArray = np.array(self.__closingPriceList, dtype=np.double)
            self.__highPriceNpArray = np.array(self.__highPriceList, dtype=np.double)
            self.__lowPriceNpArray = np.array(self.__lowPriceList, dtype=np.double)
            self.__volumeNpArray = np.array(self.__volumeList, dtype=np.double)

    def getOpeningPriceList(self):
        return self.__openingPriceList

    def getClosingPriceList(self):
        return self.__closingPriceList

    def getHighPriceList(self):
        return self.__highPriceList

    def getLowPriceList(self):
        return self.__lowPriceList

    def getVolumeList(self):
        return self.__volumeList

    def getOpeningPriceNpArray(self):
        return self.__openingPriceNpArray

    def getClosingPriceNpArray(self):
        return self.__closingPriceNpArray

    def getHighPriceNpArray(self):
        return self.__highPriceNpArray

    def getLowPriceNpArray(self):
        return self.__lowPriceNpArray

    def getVolumeNpArray(self):
        return self.__volumeNpArray

    def clear(self):
        self.__openingPriceList.clear()
        self.__closingPriceList.clear()
        self.__highPriceList.clear()
        self.__lowPriceList.clear()
        self.__volumeList.clear()
        self.__openingPriceNpArray = None
        self.__closingPriceNpArray = None
        self.__highPriceNpArray = None
        self.__lowPriceNpArray = None
        self.__volumeNpArray = None
